from agox.main import State


class FakeObserver:
    def __init__(self):
        self.set_values = ["candidates"]
        self.get_values = ["candidates"]


def test_empty_cache():
    state = State()

    # The state starts with an empty cache.
    assert state.cache == {}

    # The state can add to the cache.
    state.add_to_cache(FakeObserver(), "candidates", [], "w")

    # The state can get from the cache.
    assert state.get_from_cache(FakeObserver(), "candidates") == []



import pytest
from ase.calculators.singlepoint import SinglePointCalculator

from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.postprocessors import ParallelRelaxPostprocess, RelaxPostprocess


def make_model(environment, dataset):
    model = GPR(
        descriptor=Fingerprint(environment=environment),
        kernel=C() * RBF() + Noise(0.01),
        database=None,
    )
    model.train(dataset)
    return model


@pytest.mark.ray
@pytest.mark.parametrize(
    "postprocess_class",
    [RelaxPostprocess, ParallelRelaxPostprocess],
)
def test_postprocess_results(postprocess_class, environment_and_dataset):
    environment, dataset = environment_and_dataset
    postprocessor = postprocess_class(model=make_model(environment, dataset), optimizer_run_kwargs={"steps": 1})

    try:
        candidates = [postprocessor.postprocess(dataset[0])]
    except NotImplementedError:
        candidates = postprocessor.process_list(dataset[0:8])

    assert len(candidates) > 0

    for candidate in candidates:
        assert isinstance(candidate.calc, SinglePointCalculator)
        assert candidate.get_potential_energy() != 0



import matplotlib

matplotlib.use("Agg")

import numpy as np

from agox import AGOX
from agox.acquisitors import LowerConfidenceBoundAcquisitor
from agox.collectors import ParallelCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import SymmetryGenerator, SymmetryRattleGenerator
from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.postprocessors import ParallelRelaxPostprocess
from agox.samplers import KMeansSampler

# Manually set seed and database-index
seed = 41
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT


class RoundedEMT(EMT):
    """
    EMT calculator that rounds the energy and forces to 8 decimals.
    This is to have reproducible tests on different hardware.
    """

    def calculate(self, atoms=None, *args, **kwargs):
        super().calculate(atoms, *args, **kwargs)

        self.results["energy"] = np.round(self.results["energy"], 8)
        self.results["forces"] = np.round(self.results["forces"], 8)


calc = RoundedEMT()

##############################################################################
# System & general settings:
##############################################################################

from ase.build import fcc100

template = fcc100("Au", size=(4, 4, 2), vacuum=10.0)

z_corner = template.positions[:, 2].max() - 1.0

confinement_cell = template.cell.copy()
confinement_cell[2, 2] = 7

confinement_corner = np.array([0, 0, z_corner])
environment = Environment(
    template=template,
    symbols="Au14",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=5)

##############################################################################
# Search Settings:
##############################################################################

# Setup a ML model.
descriptor = Fingerprint(environment=environment)
beta = 0.01
k0 = C(beta, (beta, beta)) * RBF()
k1 = C(1 - beta, (1 - beta, 1 - beta)) * RBF()
kernel = C(5000, (1, 1e5)) * (k0 + k1) + Noise(0.01, (0.01, 0.01))
model = GPR(descriptor=descriptor, kernel=kernel, database=database, prior=Repulsive(), use_ray=False)

# Sampler to choose candidates to modify using the generators.
sample_size = 10
sampler = KMeansSampler(descriptor=descriptor, database=database, sample_size=sample_size)

# Generators to produce candidates structures
random_generator = SymmetryGenerator(**environment.get_confinement(), sym_type="slab")
rattle_generator = SymmetryRattleGenerator(**environment.get_confinement(), let_lose_sym=0.5)

# Dict specificies how many candidates are created with and the dict-keys are iterations.
generators = [random_generator, rattle_generator]
num_candidates = {0: [2, 0], 2: [0, 2]}

# Collector creates a number of structures in each iteration.
collector = ParallelCollector(
    generators=generators,
    sampler=sampler,
    environment=environment,
    num_candidates=num_candidates,
    order=1,
)

# Acquisitor to choose a candidate to evaluate in the real potential.
acquisitor = LowerConfidenceBoundAcquisitor(model=model, kappa=2, order=3)

# Number of steps is very low - should be set higher for a real search!
relaxer = ParallelRelaxPostprocess(
    model=acquisitor.get_acquisition_calculator(),
    constraints=environment.get_constraints(),
    optimizer_run_kwargs={"steps": 5},
    start_relax=8,
    order=2,
)


# Evaluator to evaluate the candidates in the real potential.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "prioritized_candidates"},
    optimizer_kwargs={"logfile": None},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 0},
    constraints=environment.get_constraints(),
    store_trajectory=True,
    order=4,
)

##############################################################################
# Let get the show running!
##############################################################################

# The oder of things here does not matter. But it can be simpler to understand
# what the expected behaviour is if they are put in order.
agox = AGOX(collector, relaxer, acquisitor, evaluator, database, seed=seed)

agox.run(N_iterations=10)



import numpy as np
import pytest

from agox.acquisitors import LowerConfidenceBoundAcquisitor
from agox.models.descriptors import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.test.test_utils import get_test_data, get_test_environment, test_data_dicts
from agox.utils.numerical_derivative import atomic_numerical_derivative


@pytest.mark.ray
@pytest.mark.parametrize("test_data_dict", test_data_dicts[0:1])
def test_ray_global_gpr(test_data_dict, ray_fix, cmd_options):
    # Tolerances:
    tolerance = cmd_options["tolerance"]

    # Dataset:
    path = test_data_dict["path"]
    remove = test_data_dict["remove"]
    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    train_data = data[0:25]
    test_data = data[25:35]

    descriptor = Fingerprint(environment=environment)

    kernel = C(1) * RBF(1000000) + Noise(0.01)

    model = GPR(descriptor=descriptor, kernel=kernel, n_optimize=0, use_ray=False)

    model.train(train_data)

    kappa = 0
    acquisitor = LowerConfidenceBoundAcquisitor(model=model, kappa=kappa)
    calc = acquisitor.get_acquisition_calculator()

    def calc_lcb(atoms):
        lcb = model.predict_energy(atoms) - kappa * model.predict_uncertainty(atoms)
        return lcb

    # Numerical derivative:
    atoms = test_data[0].copy()
    F = atomic_numerical_derivative(calc_lcb, atoms)

    # Analytical derivative:
    atoms = test_data[0].copy()
    atoms.set_calculator(calc)
    F_analytical = atoms.get_forces()

    print(F.shape, F_analytical.shape)

    np.testing.assert_allclose(F, F_analytical, rtol=1e2, atol=1e-10)

    # np.testing.assert_allclose(np.linalg.norm(F), np.linalg.norm(F_analytical))



import numpy as np
from ase import Atoms

from agox.environments import Environment
from agox.generators import RandomGenerator

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au8Ni8",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

random_generator = RandomGenerator(**environment.get_confinement())

candidate = random_generator(sampler=None, environment=environment)

print(f"{candidate = }")



from agox.utils.filters.random import RandomFilter


def test_random_filter(environment_and_dataset):
    environment, dataset = environment_and_dataset
    random_filter = RandomFilter(N=100)

    # Test that the filter works
    structures, indices = random_filter.filter(dataset)
    assert len(structures) == 100

    # Test that the filter works with N > len(dataset):
    random_filter = RandomFilter(N=len(dataset) * 2)
    structures, indices = random_filter.filter(dataset)

    assert len(structures) == len(dataset)



import pytest

from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.postprocessors import CenteringPostProcess, ParallelRelaxPostprocess, RelaxPostprocess, WrapperPostprocess

from copy import deepcopy

def base_setup(environment, dataset):
    return {}


def make_model(environment, dataset):
    model = GPR(descriptor=Fingerprint(environment=environment), kernel=C() * RBF() + Noise(0.01), database=None)
    model.train(dataset)
    return {"model": model}


@pytest.mark.ray
@pytest.mark.parametrize(
    "postprocess_class, setup_kwargs, setup_func",
    [
        (CenteringPostProcess, {}, base_setup),
        (WrapperPostprocess, {}, base_setup),
        (RelaxPostprocess, {"optimizer_run_kwargs": {"steps": 10}}, make_model),
        (ParallelRelaxPostprocess, {"optimizer_run_kwargs": {"steps": 10}}, make_model),
    ],
)
def test_postprocess(postprocess_class, setup_kwargs, setup_func, environment_and_dataset):
    environment, dataset = environment_and_dataset

    dataset = deepcopy(dataset[0:8])

    postprocessor = postprocess_class(**setup_kwargs, **setup_func(environment, dataset))

    try:
        postprocessor.postprocess(dataset[0])
    except NotImplementedError:
        postprocessor.process_list(dataset[0:8])




import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase.build import fcc100

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator

# Manually set seed and database-index
seed = 42
database_index = 0

##############################################################################
# Calculator
##############################################################################

from chgnet.model import CHGNetCalculator

calc = CHGNetCalculator()


##############################################################################
# System & general settings:
##############################################################################

# Define a template surface using ASE functions.
template = fcc100("Au", size=(4, 4, 2), vacuum=5.0)
template.pbc = [True, True, False]
template.positions[:, 2] -= template.positions[:, 2].min()

# Confinement cell matches the template cell in x & y but is smaller in z.
confinement_cell = template.cell.copy()
confinement_cell[2, 2] = 4.0

# Confinement corner is at the origin of the template cell, except in z where
# it is shifted up.
z0 = template.positions[:, 2].max()
confinement_corner = np.array([0, 0, z0])  # Confinement corner is at cell origin

environment = Environment(
    template=template,
    symbols="Ni8",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
    box_constraint_pbc=[True, True, False],  # Confinement is not periodic in z
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

random_generator = RandomGenerator(
    **environment.get_confinement(),
    environment=environment,
    order=1,
    contiguous=False,
)

# Wont relax fully with steps:5 - more realistic setting would be 100+.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(random_generator, database, evaluator, seed=seed)

agox.run(N_iterations=10)



def cluster_confinement():
    import numpy as np
    from ase import Atoms

    from agox.environments import Environment

    template = Atoms("", cell=np.eye(3) * 12)
    confinement_cell = np.eye(3) * 6
    confinement_corner = np.array([3, 3, 3])
    environment = Environment(
        template=template,
        symbols="Au8Ni8",
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
    )
    environment.plot("plots/cluster_confinement.png")


def surface_film_confinement():
    import numpy as np
    from ase.build import fcc100

    from agox.environments import Environment

    # Define a template surface using ASE functions.
    template = fcc100("Au", size=(4, 4, 2), vacuum=5.0)
    template.pbc = [True, True, False]
    template.positions[:, 2] -= template.positions[:, 2].min()

    # Confinement cell matches the template cell in x & y but is smaller in z.
    confinement_cell = template.cell.copy()
    confinement_cell[2, 2] = 4.0

    # Confinement corner is at the origin of the template cell, except in z where
    # it is shifted up.
    z0 = template.positions[:, 2].max()
    confinement_corner = np.array([0, 0, z0])  # Confinement corner is at cell origin

    environment = Environment(
        template=template,
        symbols="Ni8",
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
        box_constraint_pbc=[True, True, False],  # Confinement is not periodic in z
    )

    environment.plot("plots/surface_film_confinement.png")


def surface_cluster_confinement():
    import numpy as np
    from ase.build import fcc100

    from agox.environments import Environment

    # Define a template surface using ASE functions.
    template = fcc100("Au", size=(4, 4, 2), vacuum=5.0)
    template.pbc = [True, True, False]
    template.positions[:, 2] -= template.positions[:, 2].min()

    # Confinemnet cell is a 6x6x6 cube
    confinement_cell = np.eye(3) * 6

    # Confinement cell is centered in the template cell
    z0 = template.positions[:, 2].max()
    x0 = template.cell[0, 0] / 2 - confinement_cell[0, 0] / 2
    y0 = template.cell[1, 1] / 2 - confinement_cell[1, 1] / 2
    confinement_corner = np.array([x0, y0, z0])  # Confinement corner is at cell origin

    environment = Environment(
        template=template,
        symbols="Ni8",
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
        box_constraint_pbc=[True, True, False],  # Confinement is not periodic in z
    )

    environment.plot("plots/surface_cluster_confinement.png")


def two_d_environment():
    import numpy as np
    from ase import Atoms

    from agox.environments import Environment

    template = Atoms("", cell=np.eye(3) * 12)
    confinement_cell = np.eye(3) * 6
    confinement_cell[2, 2] = 0  # Zero height of confinement cell for the third dimension.
    confinement_corner = np.array([3, 3, 6])
    environment = Environment(
        template=template,
        symbols="Au4Ni4",
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
    )

    environment.plot("plots/two_d_environment.png")


def surface_toplayer_unconstrained():
    import numpy as np
    from ase.build import fcc100
    from ase.constraints import FixAtoms

    from agox.environments import Environment

    # Define a template surface using ASE functions.
    template = fcc100("Au", size=(4, 4, 2), vacuum=5.0)
    template.pbc = [True, True, False]
    template.positions[:, 2] -= template.positions[:, 2].min()

    # Confinement cell matches the template cell in x & y but is smaller in z.
    confinement_cell = template.cell.copy()
    confinement_cell[2, 2] = 4.0

    # Confinement corner is at the origin of the template cell, except in z where
    # it is shifted up.
    z0 = template.positions[:, 2].max()
    confinement_corner = np.array([0, 0, z0])  # Confinement corner is at cell origin

    # Fix bottom layers:
    constraint = FixAtoms(mask=[atom.position[2] < 2.0 for atom in template])

    environment = Environment(
        template=template,
        symbols="Ni8",
        confinement_cell=confinement_cell,
        confinement_corner=confinement_corner,
        box_constraint_pbc=[True, True, False],  # Confinement is not periodic in z
        fix_template=False,
        constraints=[constraint],
    )

    environment.plot("plots/surface_toplayer_unc.png")


if __name__ == "__main__":
    cluster_confinement()
    surface_film_confinement()
    surface_cluster_confinement()
    two_d_environment()
    surface_toplayer_unconstrained()



import pytest

from agox.test.run_tests.run_utils import agox_test_run


@pytest.mark.skip(reason="Not reproducible on CI at the moment.")
def test_gofee_cluster_symmetry(tmp_path, cmd_options, ray_fix):
    mode = "gofee_cluster_symmetry"  # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options, full_compare=False)



import numpy as np
import pytest

from agox.analysis import SearchData
from agox.analysis.property import DescriptorProperty, ListPropertyData
from agox.models.descriptors import Fingerprint


@pytest.fixture(scope="module")
def descriptor(search_data: SearchData) -> Fingerprint:
    example_atoms = search_data.get_candidate(0, 0)
    descriptor = Fingerprint.from_atoms(example_atoms)
    return descriptor


@pytest.fixture(scope="module")
def desc_property_obj(descriptor: Fingerprint) -> DescriptorProperty:
    prop = DescriptorProperty(descriptor=descriptor)
    return prop


@pytest.fixture(scope="module")
def desc_property(search_data: SearchData, desc_property_obj: DescriptorProperty) -> ListPropertyData:
    return desc_property_obj.compute(search_data)


def test_desc_property_computes(desc_property: ListPropertyData) -> None:
    print(desc_property)


def test_desc_property_hasdata(desc_property: ListPropertyData) -> None:
    assert desc_property.data is not None


def test_desc_property_hasname(desc_property: ListPropertyData) -> None:
    assert desc_property.name is not None


def test_desc_property_hasshape(desc_property: ListPropertyData) -> None:
    assert desc_property.shape is not None


def test_desc_property_hasaxis(desc_property: ListPropertyData) -> None:
    assert desc_property.axis is not None


def test_desc_property_repr(desc_property: ListPropertyData) -> None:
    assert repr(desc_property) is not None


def test_desc_property_get_minimum(desc_property_obj: DescriptorProperty, search_data: SearchData) -> None:
    with pytest.raises(ValueError):
        desc_property_obj.get_minimum([search_data])


def test_desc_property_match(
    desc_property: DescriptorProperty, descriptor: Fingerprint, search_data: SearchData
) -> None:
    fingerprint = descriptor.get_features(search_data.get_candidate(0, 0))
    np.testing.assert_allclose(desc_property.data[0][0], fingerprint)


def test_desc_property_match_shape(desc_property: ListPropertyData) -> None:
    assert desc_property.shape == ("Restarts", "Indices [#]")



import numpy as np
import pytest

from agox.analysis.search_data import SearchData


def test_search_id(search_data: SearchData) -> None:
    identifiers = search_data.get_all_identifiers()
    assert isinstance(identifiers, list)
    assert isinstance(identifiers[0], str)


def test_search_files(search_data: SearchData) -> None:
    files = search_data.get_files()
    assert isinstance(files, list)


def test_search_num_restarts(search_data: SearchData) -> None:
    n_restarts = search_data.get_number_of_restarts()
    assert n_restarts == 5


def test_search_get_energies_shape(search_data: SearchData) -> None:
    energies = search_data.get_all_energies(fill=np.nan)
    assert energies.shape[0] == search_data.get_number_of_restarts()


def test_search_get_energies_fill(search_data: SearchData) -> None:
    energies = search_data.get_all_energies(fill=np.nan)
    energies_inf = search_data.get_all_energies(fill=np.inf)
    np.testing.assert_allclose(np.nanmin(energies), np.min(energies_inf))


def test_search_get_candidate_has_energy(search_data: SearchData) -> None:
    candidate = search_data.get_candidate(0, 0)
    assert isinstance(candidate.get_potential_energy(), float)


def test_search_get_candidate_restart_index(search_data: SearchData) -> None:
    with pytest.raises(IndexError):
        search_data.get_candidate(6, 0)


def test_search_get_candidate_iteration_index(search_data: SearchData) -> None:
    with pytest.raises(IndexError):
        search_data.get_candidate(0, 1001)


def test_search_get_best_candidates_shape(search_data: SearchData) -> None:
    candidates = search_data.get_best_candidates()
    assert len(candidates) == search_data.get_number_of_restarts()


def test_search_get_best_candidates_energy(search_data: SearchData) -> None:
    candidates = search_data.get_best_candidates()
    energies = [c.get_potential_energy() for c in candidates]


def test_search_print(search_data: SearchData) -> None:
    print(search_data)


def test_search_save_load(search_data: SearchData) -> None:
    energies_b = search_data.get_all_energies()
    search_data.save()
    search_data.load()
    energies_a = search_data.get_all_energies()

    np.testing.assert_allclose(energies_b, energies_a)



import numpy as np
from ase import Atoms

from agox.environments import Environment
from agox.models.descriptors import SOAP, Fingerprint

# using AGOX environment
environment = Environment(Atoms(cell=np.eye(3) * 10), symbols="H6C6", print_report=False)
global_descriptor = Fingerprint(environment=environment, rc1=5.0)

# using ASE atoms object
atoms = Atoms("H6C6", cell=np.eye(3) * 10)
global_descriptor = Fingerprint.from_atoms(atoms, rc1=5.0)

# using AGOX environment
environment = Environment(Atoms(), symbols="HC", print_report=False)
local_descriptor = SOAP(environment=environment, r_cut=5.0)

# using species list
local_descriptor = SOAP.from_species(["H", "C"], r_cut=5.0)



from agox.models.datasets import datasets
from agox.models.descriptors import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.utils.sparsifiers import CUR

data = datasets["Ag5O3"]
transfer_data = data[:10]
training_data = data[10:80]
validation_data = data[80:90]
test_data = data[90:]

kernel = C(1, (1, 100)) * RBF(20, (10, 30))
descriptor = SOAP.from_species(
    species=["Ag", "O"],
    r_cut=5,
    nmax=3,
    lmax=2,
    sigma=1,
    weight=True,
    periodic=False,
)
model = SparseGPR(
    kernel=kernel,
    descriptor=descriptor,
    sparsifier=CUR(1000),  # sparsification using CUR algorithm with 1000 sparse points
    noise_E=0.01,  # eV/atom
    noise_F=0.05,  # eV/Å/atom
    force_data_filter="all",
)

# one can also add transfer data e.g. from another system size to improve the model
model.add_transfer_data(transfer_data, noise=0.05)

# like for the GPR model, one can add validation data
model.add_validation_data(validation_data)

# train the model on the training data
model.train(training_data)

# save the model to a file
model.save("sgpr.h5")



import pytest

from agox.test.run_tests.run_utils import agox_test_run


def test_schnet_bh(tmp_path, cmd_options):
    pytest.skip("This test is not designed to run on the CI")
    mode = "schnet_bh"  # This determines the script that is imported.
    options = {"tolerance": {"rtol": 1e-03, "atol": 1e-03}, "create_mode": cmd_options["create_mode"]}
    agox_test_run(mode, tmp_path, options)



import numpy as np
import pytest
from ase.data import atomic_numbers
from ase.data.colors import cpk_colors, jmol_colors
from matplotlib.colors import to_rgba, to_rgba_array

from agox.candidates import StandardCandidate
from agox.utils.plot.colors import Colors


@pytest.mark.parametrize("color_palette", [None, jmol_colors, cpk_colors])
def test_init(environment_and_dataset, color_palette):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    colors = Colors(structure, color_palette)

    expected_color_palette = color_palette if color_palette is not None else jmol_colors

    np.testing.assert_allclose(colors.colors, to_rgba_array(expected_color_palette[structure.numbers]))


def test_get_indices(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    colors = Colors(structure)

    assert isinstance(colors._get_indices(None), np.ndarray)
    assert isinstance(colors._get_indices(0), np.ndarray)
    assert isinstance(colors._get_indices([0, 1]), np.ndarray)
    assert isinstance(colors._get_indices(np.array([0, 1])), np.ndarray)

    np.testing.assert_array_equal(colors._get_indices(None), np.arange(len(structure)))
    np.testing.assert_array_equal(colors._get_indices(0), [0])
    np.testing.assert_array_equal(colors._get_indices([0, 1]), [0, 1])
    np.testing.assert_array_equal(colors._get_indices(np.array([0, 1])), [0, 1])


# examples from https://matplotlib.org/stable/users/explain/colors/colors.html
@pytest.mark.parametrize(
    "color",
    [
        (0.1, 0.2, 0.5),
        (0.1, 0.2, 0.5, 0.3),
        "#0f0f0f",
        "#abc",
        "0.8",
        "b",
        "aquamarine",
        "xkcd:sky blue",
        "tab:blue",
        "C0",
    ],
)
def test_set_color(environment_and_dataset, color):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    colors = Colors(structure)
    colors.set_color(color)

    expected_color = to_rgba(color)

    for color in colors.colors:
        np.testing.assert_array_equal(color, expected_color)


def test_set_alpha(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    colors = Colors(structure)
    colors.set_alpha(0.5)

    np.testing.assert_array_equal(colors.colors[:, 3], 0.5)


def test_set_element(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    one_species = sorted(structure.symbols.species())[0]

    # using symbol input
    colors = Colors(structure)
    colors.set_element(one_species, "r")

    for atom, color in zip(structure, colors):
        expected_color = to_rgba("r" if atom.symbol == one_species else jmol_colors[atom.number])
        np.testing.assert_array_equal(color, expected_color)

    # using number input
    colors = Colors(structure)
    colors.set_element(atomic_numbers[one_species], "r")

    for atom, color in zip(structure, colors):
        expected_color = to_rgba("r" if atom.symbol == one_species else jmol_colors[atom.number])
        np.testing.assert_array_equal(color, expected_color)


def test_set_template(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    colors = Colors(structure)
    colors.set_template("w")

    for i, (atom, color) in enumerate(zip(structure, colors)):
        expected_color = to_rgba("w" if i in structure.template_indices else jmol_colors[atom.number])
        np.testing.assert_array_equal(color, expected_color)


def test_darken_constant(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    colors = Colors(structure)
    colors.darken(factor=0.1)

    base_colors = to_rgba_array(jmol_colors[structure.numbers])
    assert np.all(colors.colors <= base_colors)


def test_darken_threshold(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    threshold = 0.5 * structure.cell[2, 2]

    colors = Colors(structure)
    colors.darken(factor=0.1, threshold=threshold)

    for atom, color in zip(structure, colors):
        base_color = to_rgba(jmol_colors[atom.number])
        expected_darker = atom.position[2] > threshold

        if expected_darker:
            assert np.all(color <= base_color)
        else:
            np.testing.assert_allclose(color, base_color)


def test_lighten_constant(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    colors = Colors(structure)
    colors.lighten(factor=0.1)

    base_colors = to_rgba_array(jmol_colors[structure.numbers])
    assert np.all(colors.colors >= base_colors)


def test_lighten_threshold(environment_and_dataset):
    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    threshold = 0.5 * structure.cell[2, 2]

    colors = Colors(structure)
    colors.lighten(factor=0.1, threshold=threshold)

    for atom, color in zip(structure, colors):
        base_color = to_rgba(jmol_colors[atom.number])
        expected_lighter = atom.position[2] > threshold

        if expected_lighter:
            assert np.all(color >= base_color)
        else:
            np.testing.assert_allclose(color, base_color)



import numpy as np

from agox.candidates import CandidateBaseClass


def test_non_seed_script():
    np.random.seed(42)
    from agox.test.generator_tests.for_docs.nonseeded_script import candidate

    assert isinstance(candidate, list)
    assert isinstance(candidate[0], CandidateBaseClass)


def test_seed_cript():
    np.random.seed(42)
    from agox.test.generator_tests.for_docs.seeded_script import rattle_candidate

    assert isinstance(rattle_candidate, list)
    assert isinstance(rattle_candidate[0], CandidateBaseClass)



from collections.abc import Callable

import numpy as np
from ase import Atoms

from agox.environments import Environment


def test_factory_fixtures(dataset_factory: Callable[[], list[Atoms]], environment_factory: Callable[[], Environment]) -> None:
    dataset = dataset_factory()
    environment = environment_factory()

    numbers_data = np.sort(np.unique(dataset[0].get_atomic_numbers()))
    numbers_env = np.sort(np.unique(environment.get_all_numbers()))
    np.testing.assert_array_equal(numbers_data, numbers_env)



import numpy as np
import pytest

from agox.analysis.criterion import DistanceCriterion, ThresholdCriterion
from agox.analysis.criterion.base_criterion.discrete_distribution import DiscreteDistribution
from agox.analysis.property import ArrayPropertyData


@pytest.fixture(scope="module")
def threshold_criterion() -> ThresholdCriterion:
    criterion = ThresholdCriterion(threshold=1)
    return criterion


@pytest.fixture(scope="module")
def threshold_output(
    threshold_criterion: ThresholdCriterion, energy_array_prop: ArrayPropertyData
) -> DiscreteDistribution:
    return threshold_criterion(energy_array_prop)


def test_threshold_criterion(threshold_output: DiscreteDistribution) -> None:
    assert threshold_output is not None


@pytest.fixture(scope="module")
def distance_criterion() -> DistanceCriterion:
    comparate = np.atleast_2d(np.array([1]))
    criterion = DistanceCriterion(threshold=0.1, comparate=comparate)
    return criterion


@pytest.fixture(scope="module")
def distance_output(
    distance_criterion: DistanceCriterion, energy_array_prop: ArrayPropertyData
) -> DiscreteDistribution:
    return distance_criterion(energy_array_prop)


def test_distance_criterion(distance_output: DiscreteDistribution) -> None:
    assert distance_output is not None



import os
from importlib.resources import files

import numpy as np
import pytest
from ase import Atoms
from ase.build import fcc100
from ase.io import read

from agox.candidates import StandardCandidate
from agox.environments.environment import Environment
from agox.generators import RandomGenerator
from agox.samplers.fixed_sampler import FixedSampler


def nonperiodic_environment(symbols):
    template = Atoms("", cell=np.eye(3) * 25, pbc=False)
    confinement_corner = np.array([2.5, 2.5, 2.5])
    confinement_cell = np.eye(3) * 20
    environment = Environment(
        template=template,
        symbols=symbols,
        confinement_corner=confinement_corner,
        confinement_cell=confinement_cell,
    )
    return environment


def slab_environment(symbols):
    template = fcc100("Al", size=(3, 3, 2), vacuum=10.0, orthogonal=True, periodic=True)
    template.positions[:, 2] -= template.positions[:, 2].min()

    confinement_corner = [0, 0, template.positions[:, 2].max()]
    confinement_cell = template.cell.copy()
    confinement_cell[2, 2] = 4

    environment = Environment(
        template=template, symbols=symbols, confinement_corner=confinement_corner, confinement_cell=confinement_cell
    )
    return environment


non_periodic_environment_params = [(nonperiodic_environment, ["B12"]), (nonperiodic_environment, ["B12N12"])]
non_periodic_environment_param_ids = ["nonperiodic_B12", "nonperiodic_B12N12"]


@pytest.fixture(params=non_periodic_environment_params, ids=non_periodic_environment_param_ids, scope="module")
def test_environment_non_periodic(request):
    return request.param[0](*request.param[1])


slab_environment_params = [(slab_environment, ["Al9"]), (slab_environment, ["Al5O4"])]
slab_environment_param_ids = ["slab_Al9", "slab_Al5O4"]


@pytest.fixture(params=slab_environment_params, ids=slab_environment_param_ids, scope="module")
def test_environment_slab(request):
    return request.param[0](*request.param[1])


all_environment_params = slab_environment_params + non_periodic_environment_params
all_environment_param_ids = slab_environment_param_ids + non_periodic_environment_param_ids


@pytest.fixture(params=all_environment_params, ids=all_environment_param_ids, scope="module")
def test_environment(request):
    return request.param[0](*request.param[1])


@pytest.fixture(scope="module")
def random_generator():
    def make_generator(environment):
        generator = RandomGenerator(**environment.get_confinement())
        return generator

    return make_generator


@pytest.fixture(scope="module")
def random_candidate(random_generator, test_environment):
    generator = random_generator(test_environment)
    return generator(None, test_environment)[0]


@pytest.fixture(scope="module")
def sampler(random_candidate):
    return FixedSampler(random_candidate)





import pytest

from agox.utils.plot.utils import plane_to_indices


@pytest.mark.parametrize(
    "plane,expected",
    [
        ("xy", (0, 1, 2, 1)),
        ("xy+", (0, 1, 2, 1)),
        ("xy-", (0, 1, 2, -1)),
        ("xz", (0, 2, 1, -1)),
        ("xz-", (0, 2, 1, 1)),
        ("yz+", (1, 2, 0, 1)),
        ("yx", (1, 0, 2, -1)),
    ],
)
def test_plane_to_indices(plane, expected):
    indices = plane_to_indices(plane)
    assert indices == expected


@pytest.mark.parametrize("plane", [None, "", "ab", "abc", "xx", "xyz"])
def test_invalid_plane_to_indices(plane):
    with pytest.raises(ValueError, match="Invalid plane specification"):
        plane_to_indices(plane)



from setuptools import setup  # noqa
from setuptools.extension import Extension
from Cython.Build import cythonize
import numpy

extensions = [
    Extension(
        name="agox.models.GPR.priors.repulsive",
        sources=["agox/models/GPR/priors/repulsive.pyx"],
        include_dirs=[numpy.get_include()],
    ),
    Extension(
        name="agox.models.descriptors.fingerprint_cython.angular_fingerprintFeature_cy",
        sources=["agox/models/descriptors/fingerprint_cython/angular_fingerprintFeature_cy.pyx"],
        include_dirs=[numpy.get_include()],
    ),
]

setup(
    ext_modules=cythonize(extensions),
)




from typing import Dict

import numpy as np

from agox.databases import Database


def test_number_of_candidates(database_created: Database, database_loaded: Database) -> None:
    assert len(database_created) == len(database_loaded)

def test_positions(database_created: Database, database_loaded: Database) -> None:
    for db_cand, data_cand in zip(database_created.get_all_candidates(), database_loaded.get_all_candidates()):
        np.testing.assert_allclose(db_cand.positions, data_cand.positions)

def test_cell(database_created: Database, database_loaded: Database) -> None:
    for db_cand, data_cand in zip(database_created.get_all_candidates(), database_loaded.get_all_candidates()):
        np.testing.assert_allclose(db_cand.cell, data_cand.cell)

def test_pbc(database_created: Database, database_loaded: Database) -> None:
    for db_cand, data_cand in zip(database_created.get_all_candidates(), database_loaded.get_all_candidates()):
        np.testing.assert_allclose(db_cand.pbc, data_cand.pbc)

def test_meta_keys(database_created: Database, database_loaded: Database, meta_dict: Dict) -> None:
    meta_data_names = list(meta_dict.keys())
    for db_cand, data_cand in zip(database_created.get_all_candidates(), database_loaded.get_all_candidates()):
        for meta_data_name in meta_data_names:
            assert db_cand.get_meta_information(meta_data_name) is not None
            assert data_cand.get_meta_information(meta_data_name) is not None

def test_meta_values(database_created: Database, database_loaded: Database, meta_dict: Dict) -> None:
    meta_data_names = list(meta_dict.keys())
    for db_cand, data_cand in zip(database_created.get_all_candidates(), database_loaded.get_all_candidates()):
        for meta_data_name in meta_data_names:
            db_meta = db_cand.get_meta_information(meta_data_name)
            data_meta = data_cand.get_meta_information(meta_data_name)
            assert np.array(db_meta == data_meta).all()

def test_energies(database_created: Database, database_loaded: Database) -> None:
    database_created_energies = np.array(
        [atoms.get_potential_energy() for atoms in database_created.get_all_candidates()]
    )
    database_loaded_energies = np.array(
        [atoms.get_potential_energy() for atoms in database_loaded.get_all_candidates()]
    )
    assert (database_created_energies == database_loaded_energies).all()

def test_forces(database_created: Database, database_loaded: Database) -> None:
    database_created_forces = np.array([atoms.get_forces() for atoms in database_created.get_all_candidates()])
    database_loaded_forces = np.array([atoms.get_forces() for atoms in database_loaded.get_all_candidates()])
    assert (database_created_forces == database_loaded_forces).all()


    


import functools
from typing import Any, Dict, List
from uuid import uuid4

from agox.writer import Writer


class Module:
    """
    Base class for all modules.

    Modules are the building blocks of the AGOX framework. They are used to define the
    different types of operations that are performed during a global optimization
    algorithm.

    Parameters:
    -----------
    use_cache : bool
        If True, the module will use a cache to store the results of the methods.
    """

    kwargs = ["surname"]

    def __init__(self, use_cache: bool = False, surname: bool = None, verbosity: int = Writer.STANDARD) -> None:
        self.writer = Writer(verbosity=verbosity)
        self.surname = surname if surname else ""

        self.use_cache = use_cache
        self.cache_key = str(uuid4())
        self.ray_key = str(uuid4())
        self.self_synchronizing = False
        self.dynamic_attributes = []  # Attributes that can be added and removed.

    def get_dynamic_attributes(self) -> Dict:
        return {key: self.__dict__.get(key, None) for key in self.dynamic_attributes}

    def add_dynamic_attribute(self, attribute_name: str) -> None:
        self.dynamic_attributes.append(attribute_name)

    def remove_dynamic_attribute(self, attribute_name: str) -> None:
        assert attribute_name in self.__dict__.keys()
        assert attribute_name in self.dynamic_attributes
        del self.dynamic_attributes[self.dynamic_attributes.index(attribute_name)]

    @property
    def dynamic_state(self) -> bool:
        state = len(self.dynamic_attributes) > 0
        return state

    @property
    def __name__(self) -> str:
        """
        Defines the name.
        """
        if len(self.dynamic_attributes) > 0:
            last = "Dynamic"
        else:
            last = ""
        return self.name + self.surname + last

    def find_submodules(self, in_key: str = None, 
                        only_dynamic: bool = False, 
                        top_level: bool = False
                        ) -> Dict:
        if in_key is None:
            in_key = []

        submodules = {}
        for key, value in self.__dict__.items():
            if issubclass(value.__class__, Module):
                key = in_key + [key]
                if only_dynamic:
                    if value.dynamic_state:
                        submodules[tuple(key)] = value
                else:
                    submodules[tuple(key)] = value
                submodules.update(value.find_submodules(in_key=key, only_dynamic=only_dynamic))

        if top_level: # Include only the top level if two sets of key include each other.
            keys = self._top_level_sort(list(submodules.keys()))
            submodules = {key: submodules[key] for key in keys}

        return submodules
    
    def _top_level_sort(self, tuples: List) -> List:
        from itertools import groupby

        # Sorting the list by first element to prepare for groupby
        tuples.sort(key=lambda x: x[0])

        # Grouping by the first element and keeping the shortest tuple in each group
        result = [min(group, key=len) for _, group in groupby(tuples, key=lambda x: x[0])]
        return result


    def set_for_submodule(self, submodule_keys: List, value: Any) -> None:
        reference = self
        try:
            for key in submodule_keys[0:-1]:
                last_key = key
                reference = self.__dict__[key]
            reference.__dict__[submodule_keys[-1]] = value
        except KeyError:
            raise KeyError(f"{self.name}: Could not find submodule with keys: {submodule_keys} - failed at key {last_key}")

    @classmethod
    def reset_cache_key(clc, func) -> Any:
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs) -> Any:
            self.cache_key = str(uuid4())
            return func(self, *args, **kwargs)

        return wrapper



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RattleGenerator
from agox.samplers import MetropolisSampler

# Manually set seed and database-index
seed = 42
database_index = 0

##############################################################################
# Calculator
##############################################################################

from ase.calculators.vasp import Vasp

nslots = 8  # Number of cores
calc = Vasp(
    command="mpirun -np {} vasp".format(nslots),
    prec="Low",
    lreal=False,
    lwave=False,
    xc="PBE",
    nwrite=2,
    istart=1,
    icharg=1,
    ispin=1,
    voskown=1,
    lmaxmix=6,
    nelm=500,
    nelmin=4,
    ibrion=-1,
    potim=0.3,
    iwavpr=1,
    nsw=0,
    ialgo=38,
    ismear=0,
    sigma=0.1,
    lorbit=11,
)

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12, pbc=True)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au6",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=4)

##############################################################################
# Search Settings:
##############################################################################

sampler = MetropolisSampler(temperature=0.25, order=3)

rattle_generator = RattleGenerator(**environment.get_confinement(), environment=environment, sampler=sampler, order=1)

# Wont relax fully with steps:5 - more realistic setting would be 100+.
from ase.calculators.calculator import CalculationFailed
from ase.calculators.vasp import Vasp

def callback(candidate):
    if isinstance(candidate.calc, Vasp) and not candidate.calc.read_convergence():
        raise CalculationFailed('VASP: SCF not converged')

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
    check_callback=callback,
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(rattle_generator, database, sampler, evaluator, seed=seed)

agox.run(N_iterations=10)



from pathlib import Path

from click.testing import CliRunner

from agox.cli.main import main
from agox.test.test_utils import TemporaryFolder


def test_plot_cli_help() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["graph-sort", "--help"])
    assert result.exit_code == 0


def test_graph_sort_database(tmpdir: Path, database_directories: Path) -> None:
    with TemporaryFolder(tmpdir):
        runner = CliRunner()
        result = runner.invoke(main, ["graph-sort", "-d", str(database_directories[0].resolve())])
        assert result.exit_code == 0


def test_graph_sort_trajectory(tmpdir: Path, trajectory_file_path: Path) -> None:
    with TemporaryFolder(tmpdir):
        runner = CliRunner()
        result = runner.invoke(main, ["graph-sort", "-t", str(trajectory_file_path.resolve())])
        assert result.exit_code == 0



from pathlib import Path
from typing import List

from agox.analysis import SearchCollection


def test_search_collection(search_collection: SearchCollection) -> None:
    search_collection


def test_search_collection_get_searches(search_collection: SearchCollection, database_directories: List[Path]) -> None:
    searches = search_collection.get_searches()
    assert len(searches) == len(database_directories)


def test_search_collection_iter(search_collection: SearchCollection, database_directories: List[Path]) -> None:
    i = 0
    for search in search_collection:
        i += 1
    assert i == len(database_directories)


def test_search_collection_index(search_collection: SearchCollection, database_directories: List[Path]) -> None:
    for i in range(len(database_directories)):
        search = search_collection[i]
        assert search.directory == database_directories[i]



import numpy as np
from ase import Atoms
from ase.data import covalent_radii

from agox.models.descriptors import SpectralGraphDescriptor

descriptor = SpectralGraphDescriptor(environment=None)

d = covalent_radii[1] * 2

positions = np.array([[0, 0, 0], [d, 0, 0], [0, d, 0], [d, d, 0]])

atoms = Atoms("H4", positions)

D = descriptor.get_distances(atoms)

# All atoms have two perfect neighbours and one at d * sqrt(2)
assert (np.sort(D, axis=1) == np.array([0, 1, 1, np.sqrt(2)])).all()

A_compare = np.array([[1, 1, 1, 0], [1, 1, 0, 1], [1, 0, 1, 1], [0, 1, 1, 1]])

A = descriptor.get_adjacency_matrix(atoms)

assert (A == A_compare).all()

L = descriptor.get_laplacian_matrix(atoms)

L_compare = np.eye(4) * 3 - A_compare

assert (L == L_compare).all()

feature = descriptor.get_features(atoms)

assert np.allclose(feature, np.array([-1.0, 1.0, 1.0, 3.0]))




import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator

# Manually set seed and database-index
seed = 42
database_index = 0

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

# Make an empty template with a cubic cell of size 6x6x6 with periodic boundary conditions
# in all directions
template = Atoms("", cell=np.eye(3) * 6, pbc=True)

# Confinement cell matches the template cell in all dimensions and is placed at
# the origin of the template cell.
confinement_cell = template.cell.copy()
confinement_corner = np.array([0, 0, 0])

environment = Environment(
    template=template,
    symbols="Au6Ni6",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
    box_constraint_pbc=[True, True, True],  # Confinement is periodic in all directions.
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

random_generator = RandomGenerator(**environment.get_confinement(), environment=environment, order=1)

# Wont relax fully with steps:5 - more realistic setting would be 100+.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(random_generator, database, evaluator, seed=seed)

agox.run(N_iterations=10)



import numpy as np
import pytest
from ase.io import read

from agox.candidates import StandardCandidate
from agox.environments import Environment
from agox.test.test_utils import test_data_dicts


@pytest.fixture(params=test_data_dicts)
def environment_and_dataset(request):
    atoms = read(request.param["path"])
    cell = atoms.get_cell()
    corner = np.array([0, 0, 0])
    remove = request.param["remove"]
    numbers = atoms.get_atomic_numbers()[len(atoms) - remove :]

    template = read(request.param["path"])
    del template[len(template) - remove : len(template)]
    environment = Environment(template=template, numbers=numbers, confinement_cell=cell, confinement_corner=corner)

    data = read(request.param["path"], ":")
    candidates = [StandardCandidate.from_atoms(template, a) for a in data]

    return environment, candidates


def test_environment(environment_and_dataset):
    environment, dataset = environment_and_dataset

    template = environment.get_template()
    assert template is not None
    environment.set_template(template)

    numbers = environment.get_numbers()
    assert isinstance(numbers, np.ndarray)
    environment.set_numbers(numbers)

    missing = environment.get_missing_types()
    assert isinstance(missing, np.ndarray)



import pickle

import numpy as np
from ase.io import read

from agox.candidates import StandardCandidate
from agox.test.test_utils import (
    check_file_is_deleted,
    compare_candidates,
    get_test_environment,
    get_test_sampler,
    test_folder_path,
)


def create_and_compare(generator, sampler, environment, expected_candidates, tolerance, iterations=1, seed=1):
    np.random.seed(seed)

    list_of_list_of_candidates = []
    for _ in range(iterations):
        list_of_list_of_candidates.append(generator(sampler, environment))

    if expected_candidates is not None:
        for list_of_candidates, list_of_expected in zip(list_of_list_of_candidates, expected_candidates):
            for candidate, expected_candidate in zip(list_of_candidates, list_of_expected):
                assert compare_candidates(candidate, expected_candidate, tolerance)

    return list_of_list_of_candidates


def get_test_data(path, environment):
    template = environment.get_template()
    atoms_data = read(path, index=":")
    candidate_data = [StandardCandidate.from_atoms(template, atoms) for atoms in atoms_data]
    return candidate_data


def get_name(generator, dataset_name, parameter_index):
    folder = "expected_outputs/"
    name = (
        f"{test_folder_path}generator_tests/{folder}{generator.name}_data{dataset_name}_parameter{parameter_index}.pckl"
    )
    return name


def save_expected_data(data, generator, dataset_name, parameter_index):
    name = get_name(generator, dataset_name, parameter_index)
    check_file_is_deleted(name)
    with open(name, "wb") as f:
        pickle.dump(data, f)


def load_expected_data(generator, dataset_name, parameter_index):
    name = get_name(generator, dataset_name, parameter_index)
    with open(name, "rb") as f:
        data = pickle.load(f)

    return data


def generator_testing(
    generator_class,
    test_data_dict,
    generator_args,
    base_kwargs,
    update_kwargs,
    parameter_index,
    tolerance,
    seed,
    test_mode=True,
):
    # Unpack data dictionary
    path = test_data_dict["path"]
    remove = test_data_dict["remove"]
    dataset_name = test_data_dict["name"]

    # Setup required things:
    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)
    sampler = get_test_sampler(data)

    # Create the generator
    generator_kwargs = base_kwargs.copy()
    generator_kwargs.update(update_kwargs)
    generator = generator_class(*generator_args, **environment.get_confinement(), **generator_kwargs)

    # Load expected data:
    if test_mode:
        expected_data = load_expected_data(generator, dataset_name, parameter_index)
    else:
        expected_data = None

    # Run the actual test.
    output = create_and_compare(generator, sampler, environment, expected_data, tolerance, iterations=5, seed=seed)

    if not test_mode:
        save_expected_data(output, generator, dataset_name, parameter_index)

    return output



import pytest

from agox.candidates.ABC_candidate import CandidateBaseClass
from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.samplers import (
    DistanceComparator,
    FixedSampler,
    GeneticSampler,
    KernelSimSampler,
    KMeansSampler,
    MetropolisSampler,
    SpectralGraphSampler,
)
from agox.test.test_utils import TemporaryFolder


def base_setup(environment, dataset):
    def args_func(dataset):
        return []

    return {"args_func": args_func}


def kmeans_setup(environment, dataset):
    descriptor = Fingerprint(environment=environment)
    model = GPR(descriptor=descriptor, kernel=C() * RBF() + Noise(0.01), database=None, use_ray=False)

    model.train(dataset[0:10])

    def args_func(dataset):
        return []

    return {"model": model, "descriptor": descriptor, "args_func": args_func}


def kernel_sim_setup(environment, dataset):
    descriptor = Fingerprint(environment=environment)
    kernel = C() * RBF() + Noise(0.01)
    model = GPR(descriptor=descriptor, kernel=kernel, database=None, use_ray=False)

    model.train(dataset[0:10])

    def args_func(dataset):
        return []

    return {
        "model": model,
        "descriptor": descriptor,
        "args_func": args_func,
        "kernel": None,
    }


def genetic_setup(environment, dataset):
    descriptor = Fingerprint.from_atoms(dataset[0])
    comparator = DistanceComparator(descriptor, 0.05)

    def args_func(dataset):
        return [dataset]

    return {"comparator": comparator, "args_func": args_func}


def spectral_graph_setup(environment, dataset):
    descriptor = Fingerprint(environment=environment)
    model = GPR(descriptor=descriptor, kernel=C() * RBF() + Noise(0.01), database=None, use_ray=False)
    model.train(dataset[0:10])

    def args_func(dataset):
        return []

    return {"model_calculator": model, "args_func": args_func}


def fixed_sampler_setup(environment, dataset):
    def args_func(dataset):
        return []

    return {"args_func": args_func, "sample": dataset[0:10]}


@pytest.mark.parametrize(
    "sampler_class, setup_kwargs, setup_func",
    [
        [KMeansSampler, {}, kmeans_setup],
        [MetropolisSampler, {}, base_setup],
        [GeneticSampler, {}, genetic_setup],
        [SpectralGraphSampler, {}, spectral_graph_setup],
        [KernelSimSampler, {}, kernel_sim_setup],
        [FixedSampler, {}, fixed_sampler_setup],
    ],
)
def test_sampler(sampler_class, setup_kwargs, setup_func, environment_and_dataset, tmp_path):
    with TemporaryFolder(tmp_path):
        environment, dataset = environment_and_dataset

        additional_kwargs = setup_func(environment, dataset)
        args_func = additional_kwargs.pop("args_func")
        sampler = sampler_class(**setup_kwargs, **additional_kwargs)
        sampler.iteration_counter = 0

        if not sampler.initialized:
            candidate = sampler.get_parents()  # Empty sampler should return [].
            assert candidate == []

        sampler.setup(dataset, *args_func(dataset))
        member = sampler.get_parents()
        all_members = sampler.get_all_members()
        member_calc = sampler.get_parents()

        assert issubclass(member[0].__class__, CandidateBaseClass)
        assert isinstance(all_members, list)
        assert member_calc[0].get_potential_energy()
        assert len(sampler) != 0
        assert len(sampler) == len(all_members)



import numpy as np
import pytest
from ase import Atoms
from ase.calculators.emt import EMT

from agox.models import CalculatorModel


class CountingEMT(EMT):

    def __init__(self) -> None:
        super().__init__()
        self.calculate_count = 0

    def calculate(self, *args, **kwargs) -> None:
        self.calculate_count += 1
        return super().calculate(*args, *kwargs)

@pytest.fixture
def dataset() -> list[Atoms]:

    data = []
    for _ in range(10):
        atoms = Atoms('Cu5Ni5', scaled_positions=np.random.rand(10, 3), pbc=[True, True, True], cell=np.eye(3)*5)
        atoms.calc = EMT()
        atoms.get_potential_energy()
        data.append(atoms)

    return data

@pytest.fixture
def model() -> CalculatorModel:
    return CalculatorModel(calculator=CountingEMT())

def test_predict_energy(model: CalculatorModel, dataset: list[Atoms]) -> None:
    for atoms in dataset:
        assert model.predict_energy(atoms) == atoms.get_potential_energy()

def test_predict_forces(model: CalculatorModel, dataset: list[Atoms]) -> None:
    for atoms in dataset:
        assert np.all(model.predict_forces(atoms) == atoms.get_forces())

def test_predict_energy_count(model: CalculatorModel, dataset: list[Atoms]) -> None:
    for i, atoms in enumerate(dataset):
        model.predict_energy(atoms)
        assert model.calculator.calculate_count == i+1

def test_predict_forces_count(model: CalculatorModel, dataset: list[Atoms]) -> None:
    for i, atoms in enumerate(dataset):
        model.predict_forces(atoms)
        assert model.calculator.calculate_count == i+1

def test_predict_energy_forces_count(model: CalculatorModel, dataset: list[Atoms]) -> None:
    for i, atoms in enumerate(dataset):
        model.predict_energy(atoms)
        model.predict_forces(atoms)
        assert model.calculator.calculate_count == i+1

def test_not_attach_calc(model: CalculatorModel, dataset: list[Atoms]) -> None:
    atoms = dataset[0]
    atoms.calc = None
    model.predict_energy(atoms)
    assert atoms.calc is None

def test_preserve_calc(model: CalculatorModel, dataset: list[Atoms]) -> None:
    atoms = dataset[0]
    calc_mem_address = id(atoms.calc)
    model.predict_energy(atoms)
    assert id(atoms.calc) == calc_mem_address


from agox.models.datasets import datasets
from agox.models.descriptors import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.utils.filters import EnergyFilter, SparsifierFilter
from agox.utils.sparsifiers import CUR

data = datasets["Ag5O3"]
training_data = data[:80]
validation_data = data[80:90]
test_data = data[90:]

# Initialize global descriptor using the ASE atoms interface
descriptor = Fingerprint.from_atoms(training_data[0])

# Define a filter that remove high energy structures and
# if more than 50 structure are left, pick 50 among them with CUR.
data_filter = EnergyFilter() + SparsifierFilter(descriptor=descriptor, sparsifier=CUR(m_points=50))
# Initialize the kernel
kernel = C() * RBF() + Noise(0.01)

# Initialize the model
model = GPR(
    descriptor,
    kernel,
    filter=data_filter,
    prior=Repulsive(),
    use_ray=False,
)
# One can add validation data to the model
model.add_validation_data(validation_data)

model.train(training_data)

# predict energy, uncertainty and force of test data
E_pred = model.predict_energy(test_data)
F_pred = model.predict_forces(test_data)
S_pred = model.predict_uncertainty(test_data)

# save the model to a file
model.save("gpr.h5")



from typing import Callable

import numpy as np
import pytest
from ase import Atoms
from ase.calculators.emt import EMT

from agox.candidates import Candidate
from agox.environments import Environment
from agox.generators import RandomGenerator
from agox.models import CalculatorModel
from agox.postprocessors import ParallelRelaxPostprocess


@pytest.fixture
def model() -> CalculatorModel:
    return CalculatorModel(calculator=EMT())

@pytest.fixture
def relax_postprocess(model: CalculatorModel) -> ParallelRelaxPostprocess:
    relaxer = ParallelRelaxPostprocess(model=model, 
                                       optimizer_run_kwargs={'steps': 5})
    return relaxer

@pytest.fixture
def candidate_factory() -> Callable:

    template = Atoms(cell=np.eye(3)*8)

    environment = Environment(symbols='Cu8', 
                              template=template,
                              confinement_cell=np.eye(3)*10, 
                              confinement_corner=np.ones(3))    
    generator = RandomGenerator(**environment.get_confinement())

    def factory(n=1) -> list[Candidate]:
        return [generator(sampler=None, environment=environment)[0] for _ in range(n)]
    return factory

def test_message_function(capsys, candidate_factory: Callable, relax_postprocess: ParallelRelaxPostprocess) -> None:
    candidates = candidate_factory(n=2)

    def message_function(*args, **kwargs):
        return "Test message"
    
    relax_postprocess.set_actor_message_function(message_function)
    relax_postprocess.process_list(candidates)

    output = capsys.readouterr().out
    assert "Test message" in output

def test_return_trajectory(capsys, candidate_factory: Callable, relax_postprocess: ParallelRelaxPostprocess) -> None:
    candidates = candidate_factory(n=2)
    relax_postprocess.trajectory = True
    relax_postprocess.set_trajectory_function(lambda *args, **kwargs: print("Trajectory function called"))
    relax_postprocess.process_list(candidates)

    output = capsys.readouterr().out
    assert "Trajectory function called" in output

def test_return_trajectory_type(capsys, candidate_factory: Callable, relax_postprocess: ParallelRelaxPostprocess) -> None:
    candidates = candidate_factory(n=2)
    relax_postprocess.trajectory = True

    def trajectory_function(trajectories, iteration):
        assert isinstance(trajectories, list)
        assert isinstance(trajectories[0], list)
        assert isinstance(trajectories[0][0], Atoms)

    relax_postprocess.set_trajectory_function(trajectory_function)
    relax_postprocess.process_list(candidates)




import os

# Versioning
__version_info__ = (3, 10, 0)
__version__ = "{}.{}.{}".format(*__version_info__)

# Extra versioning - Mainly for managing Pypi releases.
version_extra = os.environ.get("AGOX_VERSION_EXTRA", None)
if version_extra:
    __version__ = "{}{}".format(__version__, version_extra)

try:  # When installing the package we don't need to actually import.
    from agox.module import Module  # noqa
    from agox.observer import Observer
    from agox.writer import Writer
    from agox.cli.main import main
    from agox.main.state import State
    from agox.main.agox import AGOX

    __all__ = ["Module", "Observer", "Writer", "State", "AGOX", "__version__", "main"]

except ImportError as e:
    print(e)
    pass
